from enum import Enum
from dataclasses import dataclass


class LockedStatus(Enum):
    UNLOCK = 0
    LOCK = 1


class DoorStatus(Enum):
    CLOSED = 0
    OPEN = 1


@dataclass
class LockState:
    battery: int
    door: DoorStatus
    lock: LockedStatus


class Connectivity(Enum):
    DISCONNECTED = "disconnected"
    CONNECTED = "connected"



import asyncio
import sys
from enum import Enum
from typing import Union

from collections.abc import Callable
from bleak import BleakClient
from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.device import BLEDevice
from bleak_retry_connector import establish_connection

from .model import LockState, Connectivity
from .command import GET_LOCK_STATUS_REQ, CommandBuilder, ResponseHashCode
from .parsers import (
    parse_get_lock_status_resp,
    parse_lock_status_auto_resp,
    parse_set_uid_and_sn1_resp,
    return_payload_to_hash_code,
)
import logging

_LOGGER = logging.getLogger(__name__)


WRITE_CHAR = "6E400002-B5A3-F393-E0A9-E50E24DCCA9E"
NOTIFY_CHAR = "6E400003-B5A3-F393-E0A9-E50E24DCCA9E"
LOCK_KEY_IV = bytes("5beba2290a7fc515", "utf-8")
KEEP_ALIVE_TIMEOUT = 30


class LockOperation(Enum):
    UNLOCK = "unlock"
    LOCK = "lock"


class AegisLock:
    def __init__(self, offline_key: bytes, tiny_uid: str) -> None:
        self._lock_key = LOCK_KEY_IV
        self._offline_key = offline_key
        self._update_task: asyncio.Task[None] | None = None
        self._tiny_uid = tiny_uid
        self._command_builder = CommandBuilder(self._lock_key, offline_key, tiny_uid)
        self._pending_cmd: Union[LockOperation, None] = None
        self._pending_sn1: Union[bytes, None] = None
        self._loop = asyncio._get_running_loop()
        self._keep_alive_timer: asyncio.TimerHandle | None = None
        self._device: BLEDevice | None = None
        self._client: BleakClient | None = None
        self.state: LockState | None = None
        self._callbacks: list[Callable[[LockState, Connectivity], None]] = []
        self._enable_keep_alive = False

    async def connect(self, device: BLEDevice, keep_alive=True):
        self._enable_keep_alive = keep_alive

        def _disconnect_callback(client: BleakClient):
            _LOGGER.debug(f"{device.address} has disconnected")
            self._update_state(self.state)

        _LOGGER.debug(f"connecting to device @ {device.address}")
        self._client = await establish_connection(
            BleakClient,
            device,
            device.address,
            disconnected_callback=_disconnect_callback,
        )
        self._device = device
        _LOGGER.debug(f"connected!")
        await self._client.start_notify(NOTIFY_CHAR, self._notification_cb)

        if keep_alive:
            self._schedule_update()
            self._schedule_keep_alive()
        else:
            await self.update()

    def register_callback(self, callback: Callable[[LockState, Connectivity], None]):
        _LOGGER.debug(f"registering callback")

        def unregister_callback() -> None:
            self._callbacks.remove(callback)

        self._callbacks.append(callback)
        return unregister_callback

    def _update_state(self, new_state: LockState):
        _LOGGER.debug(f"updating state")
        self.state = new_state
        if not self._callbacks:
            return

        for callback in self._callbacks:
            try:
                connectivity = (
                    Connectivity.CONNECTED
                    if self.is_connected()
                    else Connectivity.DISCONNECTED
                )
                callback(new_state, connectivity)
            except Exception as ex:  # pylint: disable=broad-except
                # log or something?
                _LOGGER.error("Error while updating state")
                _LOGGER.error(ex)
                return

    async def _notification_cb(self, _: BleakGATTCharacteristic, data: bytes):
        hex_data = data.hex()
        hash_code = return_payload_to_hash_code(hex_data)
        _LOGGER.debug(f"incoming notification")

        if hash_code == ResponseHashCode.GetLockStatus.value:
            _LOGGER.debug(f"GetLockStatus")
            self._update_state(parse_get_lock_status_resp(hex_data))

        if hash_code == ResponseHashCode.LockStatusAuto.value:
            _LOGGER.debug(f"LockStatusAuto")
            self._update_state(parse_lock_status_auto_resp(hex_data))

        if hash_code == ResponseHashCode.SetUidAndSn1.value:
            _LOGGER.debug(f"SetUidAndSn1")
            sn2 = parse_set_uid_and_sn1_resp(hex_data)
            data = b""
            cmd = (
                self._command_builder.get_aegis_local_lock_command
                if self._pending_cmd == LockOperation.LOCK
                else self._command_builder.get_aegis_local_unlock_command
            )

            data = cmd(self._pending_sn1, sn2)
            await self._client_write(data)

    async def _client_write(self, data: bytes):
        _LOGGER.debug(f"writing data to GATT")

        # Reschedule keep alive if this isn't the update task
        if self._update_task and self._update_task.done():
            self._reschedule_keep_alive()

        await self._ensure_connection()
        return await self._client.write_gatt_char(WRITE_CHAR, data, response=True)

    async def _lock_unlock(self, pending_cmd: LockOperation):
        self._pending_cmd = pending_cmd
        self._pending_sn1 = self._command_builder.generate_sn1()
        await self._client_write(
            self._command_builder.get_send_tiny_uid_and_sn1_command(self._pending_sn1)
        )

    async def lock(self):
        _LOGGER.debug(f"Locking")
        await self._lock_unlock(LockOperation.LOCK)

    async def unlock(self):
        _LOGGER.debug(f"Unlocking")
        await self._lock_unlock(LockOperation.UNLOCK)

    async def disconnect(self):
        _LOGGER.debug(f"Disconnecting device")
        self._cancel_keepalive_timer()
        await self._client.stop_notify(NOTIFY_CHAR)
        await self._client.disconnect()
        self._client = None

    def _cancel_keepalive_timer(self) -> None:
        if self._keep_alive_timer:
            self._keep_alive_timer.cancel()
            self._keep_alive_timer = None

    async def _ensure_connection(self):
        if not self._client:
            _LOGGER.error("No client available")
        if not self._client.is_connected and self._device:
            _LOGGER.debug("disconnected. attempting reconnect")
            await self.connect(self._device, self._enable_keep_alive)

    def _reschedule_keep_alive(self):
        _LOGGER.debug("rescheduling keep alive")
        self._cancel_keepalive_timer()
        self._schedule_keep_alive()

    def _schedule_update(self):
        _LOGGER.debug("scheduling next update")
        self._update_task = asyncio.create_task(self._update_data())
        _LOGGER.debug("next update scheduled")

    def _schedule_keep_alive(self):
        _LOGGER.debug("scheduling next keep alive")
        self._keep_alive_timer = self._loop.call_later(
            KEEP_ALIVE_TIMEOUT, self._run_keep_alive
        )

    def _run_keep_alive(self):
        asyncio.create_task(self._keep_alive())

    async def _keep_alive(self):
        if not self._enable_keep_alive:
            return

        _LOGGER.debug("keep alive running")

        await self._ensure_connection()
        await self.update()
        self._schedule_keep_alive()

    async def update(self):
        await self._update_data()

    async def _update_data(self):
        _LOGGER.debug("_update_data")
        await self._client_write(GET_LOCK_STATUS_REQ)

    def address(self):
        return self._device.address

    def is_connected(self):
        return self._client.is_connected



from cryptography.hazmat.primitives import padding
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.backends import default_backend
from enum import Enum
from typing import Union
import time
import os

GET_LOCK_STATUS_REQ = bytes([91, 2, 148, 0, 150, 0, 93])

class ResponseHashCode(Enum):
    DetailInfo = 1636
    GetLockStatus = 1571
    LockStatusAuto = 1648
    SetUidAndSn1 = 1572

class Command(Enum):
    DetailInfo = 'b7'
    Lock = 'ac'
    SendTinyUidAndSn1 = '95'
    Unlock = 'ad'

class CommandBuilder:
    AEGIS_SESSION_KEY_PADDING = '61622324'
    UNLOCK_COMMAND = 'OpendoorA'
    LOCK_COMMAND = 'CloseDoor'

    def __init__(self, lock_key: bytes, offline_key: bytes, tiny_uid: str):
        self.lock_key = lock_key
        self.offline_key = offline_key
        self.tiny_uid = tiny_uid

    def encrypt(self, data: bytes, aes_key: bytes, aes_iv: bytes, has_padding: bool = True):
        cipher = Cipher(algorithms.AES(aes_key), modes.CBC(aes_iv), backend=default_backend())
        encryptor = cipher.encryptor()
        padder = padding.PKCS7(128).padder() if has_padding else None
        if has_padding:
            data = padder.update(data) + padder.finalize()
        return encryptor.update(data) + encryptor.finalize()

    def decrypt(self, encrypted_data: bytes, aes_key: bytes, aes_iv: bytes):
        cipher = Cipher(algorithms.AES(aes_key), modes.CBC(aes_iv), backend=default_backend())
        decryptor = cipher.decryptor()
        unpadder = padding.PKCS7(128).unpadder()
        decrypted_data = decryptor.update(encrypted_data) + decryptor.finalize()
        return unpadder.update(decrypted_data) + unpadder.finalize()

    def get_session_key(self, encrypt_sn2: bytes, sn1: bytes):
        decrypt_with_iv_hex_str = self.decrypt(encrypt_sn2, self.offline_key, self.lock_key).hex()
        return bytes.fromhex(sn1.hex() + self.AEGIS_SESSION_KEY_PADDING + decrypt_with_iv_hex_str)

    def load_command_with_params(self, cmd: Command, data: Union[str, None] = None):
        head = '5b'
        tail = '5d'

        if data:
            new_str = f'{cmd.value}{self.get_byte_of_data(data)}{data}'
        else:
            new_str = f'{cmd.value}00'

        byte_of_data = self.get_byte_of_data(new_str)
        padded_prefix = self.pad_prefix(hex(self.get_sum_of_data(f'{byte_of_data}{new_str}'))[2:], 2)
        big_small_change = self.get_big_small_change(2, padded_prefix)

        return bytes.fromhex(f'{head}{byte_of_data}{new_str}{big_small_change}{tail}')

    def get_sum_of_data(self, data: str):
        if len(data) < 2:
            return 0
        return int(data[:2], 16) + self.get_sum_of_data(data[2:])

    def pad_prefix(self, data: str, bits: int):
        return data.rjust(bits * 2, '0')

    def get_big_small_change(self, i: int, data: str):
        if not data:
            return ''
        len_data = i * 2
        return f'{data[len_data - 2:len_data]}{data[len_data - 4:len_data - 2]}'

    def get_byte_of_data(self, data: str):
        return self.pad_prefix(hex(len(data) // 2)[2:], 1)

    def get_detail_command(self):
        return self.load_command_with_params(Command.DetailInfo)

    def get_send_tiny_uid_and_sn1_command(self, sn1: bytes):
        new_str = f'FFFF{self.tiny_uid}{self.encrypt(sn1, self.offline_key, self.lock_key).hex()}'
        return self.load_command_with_params(Command.SendTinyUidAndSn1, new_str)

    def get_aegis_local_lock_command(self, sn1: bytes, encrypt_sn2: bytes):
        session_key = self.get_session_key(encrypt_sn2, sn1)
        command_bytes = self.LOCK_COMMAND.encode('utf-8')
        data = self.encrypt(command_bytes, session_key, self.lock_key).hex()
        return self.load_command_with_params(Command.Lock, data)

    def get_aegis_local_unlock_command(self, sn1: bytes, encrypt_sn2: bytes):
        session_key = self.get_session_key(encrypt_sn2, sn1)
        command_bytes = self.UNLOCK_COMMAND.encode('utf-8')
        pad_array = bytes.fromhex('000000')
        secs_hex = format(int(time.time()), 'x')
        padded_prefix = self.pad_prefix(secs_hex, 4)
        time_array = bytes.fromhex(padded_prefix)
        data_buffer = b''.join([command_bytes, pad_array, time_array])
        data = self.encrypt(data_buffer, session_key, self.lock_key, has_padding=False).hex()
        return self.load_command_with_params(Command.Unlock, data)

    def generate_sn1(self):
        return os.urandom(6)



import binascii
from .model import LockedStatus, DoorStatus, LockState

def get_substring(s, start, end):
    return s[start:end]

def parse_int_hex(s):
    return int(s, 16)

def parse_get_lock_status_resp(hex_str):
    lock_status = LockedStatus.LOCK if parse_int_hex(get_substring(hex_str, 8, 10)) == 1 else LockedStatus.UNLOCK
    battery = min(max(0, parse_int_hex(get_substring(hex_str, 10, 12))), 100)
    z = parse_int_hex(get_substring(hex_str, 12, 14)) == 1
    door_status = DoorStatus.CLOSED if  parse_int_hex(get_substring(hex_str, 14, 16)) == 1 else DoorStatus.OPEN

    # return {"battery": battery, "doorStatus": door_status, "lockStatus": lock_status}
    return LockState(battery, door_status, lock_status)

def parse_lock_status_auto_resp(hex_str):
    lock_status = LockedStatus.LOCK if parse_int_hex(get_substring(hex_str, 8, 10)) == 1 else LockedStatus.UNLOCK
    door_status = DoorStatus.CLOSED if parse_int_hex(get_substring(hex_str, 10, 12)) == 1 else DoorStatus.OPEN
    battery = parse_int_hex(get_substring(hex_str, 12, 14))
    # unlock_type = parse_int_hex(get_substring(hex_str, 14, 16))

    # return {"battery": battery, "doorStatus": door_status, "lockStatus": lock_status, "UnlockType": unlock_type}
    return LockState(battery, door_status, lock_status)

def hash_code(s):
    h = 0
    for i in range(len(s)):
        h = (31 * h + ord(s[i])) & 0xFFFFFFFF  # Ensure it's a 32-bit integer
    return h

def return_payload_to_hash_code(s):
    return hash_code(s[4:6])

def parse_set_uid_and_sn1_resp(hex_str):
    sn2_hex = hex_str[8:40]
    sn2_bytes = binascii.unhexlify(sn2_hex)
    return sn2_bytes


def parse_get_aegis_detail_resp(hex_str):
    lock_status = parse_int_hex(hex_str[34:36])
    door_status = parse_int_hex(hex_str[36:38])
    timeout_value = hex_str[62:66]

    return {"doorStatus": door_status, "lockStatus": lock_status, "timeoutValue": timeout_value}
